@@ -54,7 +54,7 @@ Street, Fifth Floor, Boston, MA 02110-1335 USA
 #include "backup_copy.h"
 #include "backup_mysql.h"
 #include "mysqld.h"
-#include "encryption_plugin.h"
+#include "xb_plugin.h"
 #include <sstream>
 #include <sql_error.h>
 #include "page0zip.h"
@@ -524,7 +524,8 @@ bool get_mysql_vars(MYSQL *connection)
 
   if (innodb_undo_tablespaces_var)
   {
-    srv_undo_tablespaces= strtoul(innodb_undo_tablespaces_var, &endptr, 10);
+    srv_undo_tablespaces= static_cast<uint32_t>
+      (strtoul(innodb_undo_tablespaces_var, &endptr, 10));
     ut_ad(*endptr == 0);
   }
 
@@ -1850,7 +1851,7 @@ bool write_backup_config_file()
 		"innodb_log_file_size=%llu\n"
 		"innodb_page_size=%lu\n"
 		"innodb_undo_directory=%s\n"
-		"innodb_undo_tablespaces=%lu\n"
+		"innodb_undo_tablespaces=%u\n"
 		"innodb_compression_level=%u\n"
 		"%s%s\n"
 		"%s\n",
@@ -1865,7 +1866,7 @@ bool write_backup_config_file()
 			"innodb_buffer_pool_filename=" : "",
 		innobase_buffer_pool_filename ?
 			innobase_buffer_pool_filename : "",
-		encryption_plugin_get_config());
+		xb_plugin_get_config());
 		return rc;
 }
 
