@@ -22,33 +22,37 @@ public:
     spider_string *str,
     const char *name,
     uint name_length
-  ) override;
+  );
   int append_name_with_charset(
     spider_string *str,
     const char *name,
     uint name_length,
     CHARSET_INFO *name_charset
-  ) override;
+  );
   int append_escaped_name(
     spider_string *str,
     const char *name,
     uint name_length
-  ) override;
+  );
   int append_escaped_name_with_charset(
     spider_string *str,
     const char *name,
     uint name_length,
     CHARSET_INFO *name_charset
-  ) override;
+  );
   bool is_name_quote(
     const char head_code
-  ) override;
+  );
   int append_escaped_name_quote(
     spider_string *str
-  ) override;
-  int append_column_value(ha_spider *spider, spider_string *str, Field *field,
-                          const uchar *new_ptr, bool is_like,
-                          CHARSET_INFO *access_charset) override;
+  );
+  int append_column_value(
+    ha_spider *spider,
+    spider_string *str,
+    Field *field,
+    const uchar *new_ptr,
+    CHARSET_INFO *access_charset
+  );
   int append_from_with_alias(
     spider_string *str,
     const char **table_names,
@@ -62,19 +66,19 @@ public:
   int append_trx_isolation(
     spider_string *str,
     int trx_isolation
-  ) override;
+  );
   int append_autocommit(
     spider_string *str,
     bool autocommit
-  ) override;
+  );
   int append_sql_log_off(
     spider_string *str,
     bool sql_log_off
-  ) override;
+  );
   int append_wait_timeout(
     spider_string *str,
     int wait_timeout
-  ) override;
+  );
   virtual int append_sql_mode_internal(
     spider_string *str,
     sql_mode_t sql_mode
@@ -82,21 +86,25 @@ public:
   int append_sql_mode(
     spider_string *str,
     sql_mode_t sql_mode
-  ) override;
+  );
   int append_time_zone(
     spider_string *str,
     Time_zone *time_zone
-  ) override;
+  );
+  int append_loop_check(
+    spider_string *str,
+    SPIDER_CONN *conn
+  );
   int append_start_transaction(
     spider_string *str
-  ) override;
+  );
   int append_xa_start(
     spider_string *str,
     XID *xid
-  ) override;
+  );
   int append_lock_table_head(
     spider_string *str
-  ) override;
+  );
   int append_lock_table_body(
     spider_string *str,
     const char *db_name,
@@ -106,13 +114,13 @@ public:
     uint table_name_length,
     CHARSET_INFO *table_name_charset,
     int lock_type
-  ) override;
+  );
   int append_lock_table_tail(
     spider_string *str
-  ) override;
+  );
   int append_unlock_table(
     spider_string *str
-  ) override;
+  );
   int open_item_func(
     Item_func *item_func,
     ha_spider *spider,
@@ -121,7 +129,7 @@ public:
     uint alias_length,
     bool use_fields,
     spider_fields *fields
-  ) override;
+  );
 #ifdef HANDLER_HAS_DIRECT_AGGREGATE
   int open_item_sum_func(
     Item_sum *item_sum,
@@ -131,12 +139,12 @@ public:
     uint alias_length,
     bool use_fields,
     spider_fields *fields
-  ) override;
+  );
 #endif
   int append_escaped_util(
     spider_string *to,
     String *from
-  ) override;
+  );
 #ifdef SPIDER_HAS_GROUP_BY_HANDLER
   int append_table(
     ha_spider *spider,
@@ -178,19 +186,26 @@ public:
     spider_string *str,
     TABLE_LIST *table_list,
     uint table_count
-  ) override;
+  );
   int reappend_tables(
     spider_fields *fields,
     SPIDER_LINK_IDX_CHAIN *link_idx_chain,
     spider_string *str
-  ) override;
+  );
   int append_where(
     spider_string *str
-  ) override;
+  );
   int append_having(
     spider_string *str
-  ) override;
+  );
 #endif
+  bool tables_on_different_db_are_joinable();
+  bool socket_has_default_value();
+  bool database_has_default_value();
+  bool default_file_has_default_value();
+  bool host_has_default_value();
+  bool port_has_default_value();
+  bool append_charset_name_before_string();
 };
 
 class spider_db_mysql_util: public spider_db_mbase_util
@@ -203,9 +218,8 @@ public:
     spider_string *str,
     Field *field,
     const uchar *new_ptr,
-    bool is_like,
     CHARSET_INFO *access_charset
-  ) override;
+  );
 };
 
 class spider_db_mariadb_util: public spider_db_mbase_util
@@ -216,15 +230,14 @@ public:
   int append_sql_mode_internal(
     spider_string *str,
     sql_mode_t sql_mode
-  ) override;
+  );
   int append_column_value(
     ha_spider *spider,
     spider_string *str,
     Field *field,
     const uchar *new_ptr,
-    bool is_like,
     CHARSET_INFO *access_charset
-  ) override;
+  );
 };
 
 class spider_db_mbase_row: public spider_db_row
@@ -437,7 +450,7 @@ public:
   bool is_xa_nota_error(
     int error_num
   );
-  void print_warnings(
+  int print_warnings(
     struct tm *l_time
   );
   spider_db_result *store_result(
@@ -446,6 +459,7 @@ public:
     int *error_num
   );
   spider_db_result *use_result(
+    ha_spider *spider,
     st_spider_db_request_key *request_key,
     int *error_num
   );
@@ -527,6 +541,11 @@ public:
     Time_zone *time_zone,
     int *need_mon
   );
+  bool set_loop_check_in_bulk_sql();
+  int set_loop_check(
+    int *need_mon
+  );
+  int fin_loop_check();
   int exec_simple_sql_with_result(
     SPIDER_TRX *trx,
     SPIDER_SHARE *share,
@@ -560,54 +579,6 @@ public:
     uint binlog_pos_length,
     SPIDER_DB_RESULT **res
   );
-#if defined(HS_HAS_SQLCOM) && defined(HAVE_HANDLERSOCKET)
-  int append_sql(
-    char *sql,
-    ulong sql_length,
-    st_spider_db_request_key *request_key
-  );
-  int append_open_handler(
-    uint handler_id,
-    const char *db_name,
-    const char *table_name,
-    const char *index_name,
-    const char *sql,
-    st_spider_db_request_key *request_key
-  );
-  int append_select(
-    uint handler_id,
-    spider_string *sql,
-    SPIDER_DB_HS_STRING_REF_BUFFER *keys,
-    int limit,
-    int skip,
-    st_spider_db_request_key *request_key
-  );
-  int append_insert(
-    uint handler_id,
-    SPIDER_DB_HS_STRING_REF_BUFFER *upds,
-    st_spider_db_request_key *request_key
-  );
-  int append_update(
-    uint handler_id,
-    spider_string *sql,
-    SPIDER_DB_HS_STRING_REF_BUFFER *keys,
-    SPIDER_DB_HS_STRING_REF_BUFFER *upds,
-    int limit,
-    int skip,
-    bool increment,
-    bool decrement,
-    st_spider_db_request_key *request_key
-  );
-  int append_delete(
-    uint handler_id,
-    spider_string *sql,
-    SPIDER_DB_HS_STRING_REF_BUFFER *keys,
-    int limit,
-    int skip,
-    st_spider_db_request_key *request_key
-  );
-  void reset_request_queue();
-#endif
   size_t escape_string(
     char *to,
     const char *from,
@@ -795,9 +766,6 @@ protected:
   spider_string           *exec_ha_sql;
   bool                    reading_from_bulk_tmp_table;
   bool                    filled_up;
-#if defined(HS_HAS_SQLCOM) && defined(HAVE_HANDLERSOCKET)
-  SPIDER_DB_HS_STRING_REF_BUFFER hs_upds;
-#endif
   SPIDER_INT_HLD          *union_table_name_pos_first;
   SPIDER_INT_HLD          *union_table_name_pos_current;
 public:
@@ -904,14 +872,6 @@ public:
   int append_delete(
     spider_string *str
   );
-  #if defined(HS_HAS_SQLCOM) && defined(HAVE_HANDLERSOCKET)
-  #ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
-  int append_increment_update_set_part();
-  int append_increment_update_set(
-    spider_string *str
-  );
-  #endif
-  #endif
   int append_update_set_part();
   int append_update_set(
     spider_string *str
@@ -1379,15 +1339,15 @@ public:
     int link_idx
   );
   int append_explain_select_part(
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     ulong sql_type,
     int link_idx
   );
   int append_explain_select(
     spider_string *str,
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     ulong sql_type,
     int link_idx
   );
@@ -1442,23 +1402,6 @@ public:
   int reset_sql(
     ulong sql_type
   );
-#if defined(HS_HAS_SQLCOM) && defined(HAVE_HANDLERSOCKET)
-  int reset_keys(
-    ulong sql_type
-  );
-  int reset_upds(
-    ulong sql_type
-  );
-  int reset_strs(
-    ulong sql_type
-  );
-  int reset_strs_pos(
-    ulong sql_type
-  );
-  int push_back_upds(
-    SPIDER_HS_STRING_REF &info
-  );
-#endif
   bool need_lock_before_set_sql_for_exec(
     ulong sql_type
   );
@@ -1516,8 +1459,8 @@ public:
     ulonglong &last_insert_id
   );
   ha_rows explain_select(
-    key_range *start_key,
-    key_range *end_key,
+    const key_range *start_key,
+    const key_range *end_key,
     int link_idx
   );
   int lock_tables(
@@ -1661,6 +1604,18 @@ public:
     spider_fields *fields
   );
 #endif
+#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
+  bool check_direct_update(
+    st_select_lex *select_lex,
+    longlong select_limit,
+    longlong offset_limit
+  );
+  bool check_direct_delete(
+    st_select_lex *select_lex,
+    longlong select_limit,
+    longlong offset_limit
+  );
+#endif
 };
 
 class spider_mysql_handler: public spider_mbase_handler
